import matplotlib.pyplot as plt
import numpy as np
from dataclasses import dataclass
import scipy.interpolate as interpolate

from .blocks import *


@dataclass
class Ring:
    """Ring object to build meshes."""

    blocks: list
    surf_top: list
    surf_bt: list
    surf_rad: list
    _grading_r: str = "1"
    _grading_z: str = "1"

    def _update_grading(self):
        for b in self.blocks:
            b.grading = f"simpleGrading ({self._grading_r} 1 {self._grading_z})"

    def set_grading_radial(self, val):
        self._grading_r = val
        self._update_grading()

    def set_grading_axial(self, val):
        self._grading_z = val
        self._update_grading()

    def set_spline_surface(self, spline, pos, res=100):
        # set surface with spline z=f(r)

        # TODO update inside and outside points
        # if type(z_bt) is interp1d:
        #     z_bt_out = z_bt(r_bt)
        # if type(z_top) is interp1d:
        #     z_top_out = z_top(r_top)

        if pos == "bottom":
            r_in = (self.blocks[0]._p0.x1 ** 2 + self.blocks[0]._p0.x2 ** 2) ** 0.5
            r_out = (self.blocks[0]._p1.x1 ** 2 + self.blocks[0]._p1.x2 ** 2) ** 0.5
            self.blocks[0].p0.update_coordinates(cartesian(r_in, 0, spline(r_in)))
            self.blocks[1].p0.update_coordinates(cartesian(r_in, 90, spline(r_in)))
            self.blocks[2].p0.update_coordinates(cartesian(r_in, 180, spline(r_in)))
            self.blocks[3].p0.update_coordinates(cartesian(r_in, 270, spline(r_in)))
            self.blocks[0].p1.update_coordinates(cartesian(r_out, 0, spline(r_out)))
            self.blocks[1].p1.update_coordinates(cartesian(r_out, 90, spline(r_out)))
            self.blocks[2].p1.update_coordinates(cartesian(r_out, 180, spline(r_out)))
            self.blocks[3].p1.update_coordinates(cartesian(r_out, 270, spline(r_out)))
            for b in self.blocks:
                b.e0.type = "spline"
            for r in np.linspace(r_in, r_out, res, endpoint=False):
                self.blocks[0].e0.points.append(cartesian(r, 0, spline(r)))
                self.blocks[1].e0.points.append(cartesian(r, 90, spline(r)))
                self.blocks[2].e0.points.append(cartesian(r, 180, spline(r)))
                self.blocks[3].e0.points.append(cartesian(r, 270, spline(r)))
        elif pos == "top":
            r_in = (self.blocks[0]._p4.x1 ** 2 + self.blocks[0]._p4.x2 ** 2) ** 0.5
            r_out = (self.blocks[0]._p5.x1 ** 2 + self.blocks[0]._p5.x2 ** 2) ** 0.5
            self.blocks[0].p4.update_coordinates(cartesian(r_in, 0, spline(r_in)))
            self.blocks[1].p4.update_coordinates(cartesian(r_in, 90, spline(r_in)))
            self.blocks[2].p4.update_coordinates(cartesian(r_in, 180, spline(r_in)))
            self.blocks[3].p4.update_coordinates(cartesian(r_in, 270, spline(r_in)))
            self.blocks[0].p5.update_coordinates(cartesian(r_out, 0, spline(r_out)))
            self.blocks[1].p5.update_coordinates(cartesian(r_out, 90, spline(r_out)))
            self.blocks[2].p5.update_coordinates(cartesian(r_out, 180, spline(r_out)))
            self.blocks[3].p5.update_coordinates(cartesian(r_out, 270, spline(r_out)))
            for b in self.blocks:
                b.e3.type = "spline"
            for r in np.linspace(r_in, r_out, res, endpoint=False):
                self.blocks[0].e3.points.append(cartesian(r, 0, spline(r)))
                self.blocks[1].e3.points.append(cartesian(r, 90, spline(r)))
                self.blocks[2].e3.points.append(cartesian(r, 180, spline(r)))
                self.blocks[3].e3.points.append(cartesian(r, 270, spline(r)))
        elif pos == "side":
            r_bt = (self.blocks[0]._p1.x1 ** 2 + self.blocks[0]._p1.x2 ** 2) ** 0.5
            r_top = (self.blocks[0]._p5.x1 ** 2 + self.blocks[0]._p5.x2 ** 2) ** 0.5
            self.blocks[0].p1.update_coordinates(cartesian(r_bt, 0, spline(r_bt)))
            self.blocks[1].p1.update_coordinates(cartesian(r_bt, 90, spline(r_bt)))
            self.blocks[2].p1.update_coordinates(cartesian(r_bt, 180, spline(r_bt)))
            self.blocks[3].p1.update_coordinates(cartesian(r_bt, 270, spline(r_bt)))
            self.blocks[0].p5.update_coordinates(cartesian(r_top, 0, spline(r_top)))
            self.blocks[1].p5.update_coordinates(cartesian(r_top, 90, spline(r_top)))
            self.blocks[2].p5.update_coordinates(cartesian(r_top, 180, spline(r_top)))
            self.blocks[3].p5.update_coordinates(cartesian(r_top, 270, spline(r_top)))
            for b in self.blocks:
                b.e9.type = "spline"
            for r in np.linspace(r_bt, r_top, res):
                self.blocks[0].e9.points.append(cartesian(r, 0, spline(r)))
                self.blocks[1].e9.points.append(cartesian(r, 90, spline(r)))
                self.blocks[2].e9.points.append(cartesian(r, 180, spline(r)))
                self.blocks[3].e9.points.append(cartesian(r, 270, spline(r)))
        else:
            raise ValueError(
                "This position does not exists. Allowable parameters for pos are 'bottom', 'top', 'side'."
            )


@dataclass
class Cylinder:
    """Cylinder object to build meshes."""

    core: Block
    ring: Ring
    surf_top: list
    surf_bt: list
    surf_rad: list
    _grading_r: str = "1"
    _grading_z: str = "1"

    def _update_grading(self):
        self.core.grading = f"simpleGrading (1 1 {self._grading_z})"
        self.ring.set_grading_radial(self._grading_r)
        self.ring.set_grading_axial(self._grading_z)

    def set_grading_radial(self, val):
        self._grading_r = val
        self._update_grading()

    def set_grading_axial(self, val):
        self._grading_z = val
        self._update_grading()

    def set_spline_surface(self, spline, pos, res=100):
        # set surface with spline z=f(r)

        if pos == "bottom":
            # update core
            radius_center_bt = (self.core._p1.x1 ** 2 + self.core._p1.x2 ** 2) ** 0.5
            z_bt_mid = spline(radius_center_bt)
            # update points
            self.core.p0.update_coordinates(cartesian(radius_center_bt, 0, z_bt_mid))
            self.core.p1.update_coordinates(cartesian(radius_center_bt, 90, z_bt_mid))
            self.core.p2.update_coordinates(cartesian(radius_center_bt, 180, z_bt_mid))
            self.core.p3.update_coordinates(cartesian(radius_center_bt, 270, z_bt_mid))
            # set values on edges
            x0 = np.array(cartesian(radius_center_bt, 0, z_bt_mid))
            x1 = np.array(cartesian(radius_center_bt, 90, z_bt_mid))
            radii = []
            phis = []
            for dist in np.linspace(0, 1, res, endpoint=False):
                pos = x0 + dist * (x1 - x0)
                radii.append((pos[0] ** 2 + pos[1] ** 2) ** 0.5)
                phis.append(np.arctan(pos[1] / pos[0]) * 360 / (2 * np.pi))
            z_vals = spline(radii)
            self.core.e0.type = "spline"
            self.core.e5.type = "spline"
            self.core.e1.type = "spline"
            self.core.e4.type = "spline"
            for i in range(len(radii) - 1):
                self.core.e0.points.append(
                    cartesian(radii[i + 1], phis[i + 1], z_vals[i + 1])
                )
                self.core.e5.points.append(
                    cartesian(radii[i + 1], phis[i + 1] + 90, z_vals[i + 1])
                )
                self.core.e1.points.append(
                    cartesian(radii[i + 1], 270 - phis[i + 1], z_vals[i + 1])
                )
                self.core.e4.points.append(
                    cartesian(radii[i + 1], 360 - phis[i + 1], z_vals[i + 1])
                )
            self.ring.set_spline_surface(spline, "bottom", res)

        elif pos == "top":
            # update core
            radius_center_top = (self.core._p5.x1 ** 2 + self.core._p5.x2 ** 2) ** 0.5
            z_top_mid = spline(radius_center_top)
            # update points
            self.core.p4.update_coordinates(cartesian(radius_center_top, 0, z_top_mid))
            self.core.p5.update_coordinates(cartesian(radius_center_top, 90, z_top_mid))
            self.core.p6.update_coordinates(
                cartesian(radius_center_top, 180, z_top_mid)
            )
            self.core.p7.update_coordinates(
                cartesian(radius_center_top, 270, z_top_mid)
            )
            # set values on edges
            x0 = np.array(cartesian(radius_center_top, 0, z_top_mid))
            x1 = np.array(cartesian(radius_center_top, 90, z_top_mid))
            radii = []
            phis = []
            for dist in np.linspace(0, 1, res, endpoint=False):
                pos = x0 + dist * (x1 - x0)
                radii.append((pos[0] ** 2 + pos[1] ** 2) ** 0.5)
                phis.append(np.arctan(pos[1] / pos[0]) * 360 / (2 * np.pi))
            z_vals = spline(radii)

            self.core.e3.type = "spline"
            self.core.e6.type = "spline"
            self.core.e2.type = "spline"
            self.core.e7.type = "spline"

            for i in range(len(radii) - 1):
                self.core.e3.points.append(
                    cartesian(radii[i + 1], phis[i + 1], z_vals[i + 1])
                )
                self.core.e6.points.append(
                    cartesian(radii[i + 1], phis[i + 1] + 90, z_vals[i + 1])
                )
                self.core.e2.points.append(
                    cartesian(radii[i + 1], 270 - phis[i + 1], z_vals[i + 1])
                )
                self.core.e7.points.append(
                    cartesian(radii[i + 1], 360 - phis[i + 1], z_vals[i + 1])
                )
            self.ring.set_spline_surface(spline, "top", res)

        elif pos == "side":
            raise NotImplementedError()
        else:
            raise ValueError(
                "This position does not exists. Allowable parameters for pos are 'bottom', 'top', 'side'."
            )


def cartesian(r, phi, z, degree=True):
    """Convert cylindrical to cartesian coordinates.

    Args:
        r (float): Radius.
        phi (float): Angle in degree.
        z (float): Axial coordinate.
        degree (bool): If false: use radiant.

    Returns:
        cartesian coordinates (x, y, z)
    """
    if degree:
        phi = 2 * np.pi * phi / 360
    x = r * np.cos(phi)
    y = r * np.sin(phi)
    return [x, y, z]


def spline(points, kind="cubic"):
    """Create a spline functions

    Args:
        points (list): list of [r, z] coordinates
        kind (str, optional): spline type (linear / cubic)
    """
    points = np.array(points)
    return interpolate.interp1d(points[:, 0], points[:, 1], kind=kind)


# inspired by https://github.com/kawache/Python-B-spline-examples
def b_spline(control_points, plot=False, res=100):
    """Create a spline curve that does not go through the control points.

    Args:
        control_points (list): list of points [[x0, y0], [x1, y1], ...]
        plot (bool): Plot the spline and the control points
        res (int): Resolution. Defaults to 100.

    Returns:
        array: [x coordinates, y coordinates]
    """
    control_points = np.array(control_points)
    t = np.concatenate(
        [[0, 0, 0], np.linspace(0, 1, control_points.shape[0] - 2), [1, 1, 1]]
    )  # knots

    spline = interpolate.splev(np.linspace(0, 1, res), [t, control_points.T, 3])

    if plot:
        fig, ax = plt.subplots(1, 1)
        ax.plot(spline[0], spline[1], label="b-spline curve")
        ax.plot(
            control_points[:, 0], control_points[:, 1], "x--", label="control points"
        )
        ax.legend()
        ax.grid(linestyle=":")
        plt.show()
    return np.array(spline)


def plot_spline(sp, r, fig=None, ax=None):
    """plot spline function

    Args:
        sp (function): spline function
        r (list): [min, max radius]
        fig (figure, optional): Matplotlib figure.
        ax (axis, optional): Matplotlib axis.

    Returns:
        matplotlib figure, axis
    """
    if fig is None:
        fig, ax = plt.subplots(1, 1)
    r = np.linspace(r[0], r[1], 100)
    ax.plot(r, sp(r))
    return fig, ax


def create_cylinder(
    mesh,
    p_top,
    p_bt,
    res_r,
    res_phi,
    res_z,
    radius_ratio=0.5,
    cylinder_below=None,
    cylinder_on_top=None,
):
    """Create a cylinder object consisting of a square core surrounded by a ring.

    Args:
        mesh (Mesh): Mesh containing the blocks.
        p_top (list): [r, z] coordinate at top
        p_bt (list): [r, z] coordinate at bottom
        res_r (float): Number of cells in radial direction (in Ring).
        res_phi (float): Number of cells in circumferential direction.
        res_z (float): Number of cells in axial direction.
        radius_ratio (float, optional): Fraction inner square / total radius.
        cylinder_below (Cylinder, optional): Cylinder object below this cylinder.
        cylinder_on_top (Cylinder, optional): Cylinder object on top of this cylinder.

    Returns:
        Cylinder object.
    """
    if cylinder_below is not None and cylinder_on_top is not None:
        raise ValueError("Both cylinder on top and below is not supported.")

    # Central block
    radius_center_top = p_top[0] * radius_ratio
    radius_center_bt = p_bt[0] * radius_ratio
    res_center_r = int(round(res_phi / 4))

    if cylinder_below is not None:
        b = Block(mesh)
        b.set_connection(cylinder_below.core, "bottom")
        b.p4 = cartesian(radius_center_top, 0, p_top[1])
        b.p5 = cartesian(radius_center_top, 90, p_top[1])
        b.p6 = cartesian(radius_center_top, 180, p_top[1])
        b.p7 = cartesian(radius_center_top, 270, p_top[1])
    elif cylinder_on_top is not None:
        b = Block(mesh)
        b.set_connection(cylinder_on_top.core, "top")
        b.p0 = cartesian(radius_center_bt, 0, p_bt[1])
        b.p1 = cartesian(radius_center_bt, 90, p_bt[1])
        b.p2 = cartesian(radius_center_bt, 180, p_bt[1])
        b.p3 = cartesian(radius_center_bt, 270, p_bt[1])
    else:
        b = Block(
            mesh,
            cartesian(radius_center_bt, 0, p_bt[1]),
            cartesian(radius_center_bt, 90, p_bt[1]),
            cartesian(radius_center_bt, 180, p_bt[1]),
            cartesian(radius_center_bt, 270, p_bt[1]),
            cartesian(radius_center_top, 0, p_top[1]),
            cartesian(radius_center_top, 90, p_top[1]),
            cartesian(radius_center_top, 180, p_top[1]),
            cartesian(radius_center_top, 270, p_top[1]),
        )
    b.set_number_of_cells(res_center_r, res_center_r, res_z)
    b.create()

    # create ring around these blocks
    ring_below = None
    ring_on_top = None
    if cylinder_below is not None:
        ring_below = cylinder_below.ring
    if cylinder_on_top is not None:
        ring_on_top = cylinder_on_top.ring
    ring = create_ring(
        mesh,
        p_top,
        p_bt,
        [b.face_front, b.face_right, b.face_back, b.face_left],
        res_r,
        res_phi,
        res_z,
        ring_below=ring_below,
        ring_on_top=ring_on_top,
    )
    surf_top = [b.face_top] + ring.surf_top
    surf_bt = [b.face_bottom] + ring.surf_bt

    return Cylinder(b, ring, surf_top, surf_bt, ring.surf_rad)


def create_ring(
    mesh,
    p_top,
    p_bt,
    faces_inside,
    res_r,
    res_phi,
    res_z,
    ring_below=None,
    ring_on_top=None,
    faces_outside=[],
):
    """Create a ring of 4 blocks.

    Args:
        mesh (Mesh): Mesh object containing the blocks.
        p_top (list): [r, z] coordinate at top
        p_bt (list): [r, z] coordinate at bottom
        faces_inside (list): Faces of blocks inside of the ring.
        res_r (float): Number of cells in radial direction.
        res_phi (float): Number of cells in circumferential direction.
        res_z (float): Number of cells in axial direction.
        ring_below (Ring, optional): Ring object bellow this ring.
        ring_on_top (Ring, optional): Ring object on top of this ring.
        faces_outside (list, optional): Surfaces of blocks outside of this ring.

    Returns:
        Ring object.
    """

    if ring_below is not None and ring_on_top is not None:
        raise ValueError("It's not allowed to provide both ring_on_top and ring_below.")

    blocks = []

    b = Block(mesh)
    b.face_left = faces_inside[0]
    if faces_outside != []:
        b.face_right = faces_outside[0]
        b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
        b.create()
    elif ring_below is not None:
        b.set_connection(ring_below.blocks[0], "bottom")
        b.p5 = cartesian(p_top[0], 0, p_top[1])
        b.p6 = cartesian(p_top[0], 90, p_top[1])
        b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
        b.create()
        b.e6.type = "arc"
        b.e6.points = [cartesian(p_top[0], 45, p_top[1])]
    elif ring_on_top is not None:
        b.set_connection(ring_on_top.blocks[0], "top")
        b.p1 = cartesian(p_bt[0], 0, p_bt[1])
        b.p2 = cartesian(p_bt[0], 90, p_bt[1])
        b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
        b.create()
        b.e5.type = "arc"
        b.e5.points = [cartesian(p_bt[0], 45, p_bt[1])]
    else:
        b.p1 = cartesian(p_bt[0], 0, p_bt[1])
        b.p2 = cartesian(p_bt[0], 90, p_bt[1])
        b.p5 = cartesian(p_top[0], 0, p_top[1])
        b.p6 = cartesian(p_top[0], 90, p_top[1])
        b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
        b.create()
        b.e5.type = "arc"
        b.e5.points = [cartesian(p_bt[0], 45, p_bt[1])]
        b.e6.type = "arc"
        b.e6.points = [cartesian(p_top[0], 45, p_top[1])]
    blocks.append(b)

    for i in range(2):
        i += 1
        b = Block(mesh)
        b.face_left = faces_inside[i]
        b.face_front = blocks[i - 1].face_back
        if faces_outside != []:
            b.face_right = faces_outside[i]
            b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
            b.create()
        elif ring_below is not None:
            b.set_connection(ring_below.blocks[i], "bottom")
            b.p6 = cartesian(p_top[0], 90 * (i + 1), p_top[1])
            b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
            b.create()
            b.e6.type = "arc"
            b.e6.points = [cartesian(p_top[0], i * 90 + 45, p_top[1])]
        elif ring_on_top is not None:
            b.set_connection(ring_on_top.blocks[i], "top")
            b.p2 = cartesian(p_bt[0], 90 * (i + 1), p_bt[1])
            b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
            b.create()
            b.e5.type = "arc"
            b.e5.points = [cartesian(p_bt[0], i * 90 + 45, p_bt[1])]
        else:
            b.p2 = cartesian(p_bt[0], 90 * (i + 1), p_bt[1])
            b.p6 = cartesian(p_top[0], 90 * (i + 1), p_top[1])
            b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
            b.create()
            b.e5.type = "arc"
            b.e5.points = [cartesian(p_bt[0], i * 90 + 45, p_bt[1])]
            b.e6.type = "arc"
            b.e6.points = [cartesian(p_top[0], i * 90 + 45, p_top[1])]
        blocks.append(b)

    b = Block(mesh)
    b.face_left = faces_inside[-1]
    b.face_front = blocks[-1].face_back
    b.face_back = blocks[0].face_front
    b.set_number_of_cells(res_r, int(round(res_phi / 4)), res_z)
    b.create()
    if faces_outside != []:
        pass
    elif ring_below is not None:
        b.e6.type = "arc"
        b.e6.points = [cartesian(p_top[0], 360 - 45, p_top[1])]
    elif ring_on_top is not None:
        b.e5.type = "arc"
        b.e5.points = [cartesian(p_bt[0], 360 - 45, p_bt[1])]
    else:
        b.e5.type = "arc"
        b.e5.points = [cartesian(p_bt[0], 360 - 45, p_bt[1])]
        b.e6.type = "arc"
        b.e6.points = [cartesian(p_top[0], 360 - 45, p_top[1])]
    blocks.append(b)

    surf_top = []
    surf_bt = []
    surf_rad = []
    for b in blocks:
        surf_top.append(b.face_top)
        surf_bt.append(b.face_bottom)
        surf_rad.append(b.face_right)

    return Ring(blocks, surf_top, surf_bt, surf_rad)



from dataclasses import dataclass
import numpy as np
import os


class Mesh:
    """Collection of everything that belongs into the blockMeshDict."""

    def __init__(self) -> None:
        self.blocks = []
        self.block_count = 0

        self.points = []
        self.point_count = 0

        self.edges = []

        self.patches = []

    def _add_point(self, x1, x2, x3):
        p = Point(x1, x2, x3, self.point_count)
        self.point_count += 1
        self.points.append(p)
        return p

    def _add_edge(self, p0, p1):
        e = Edge(p0, p1)
        self.edges.append(e)
        return e

    def _add_block(self, block):
        block.id = self.block_count
        self.block_count += 1
        self.blocks.append(block)

    def write(self):
        if not os.path.exists("./system"):
            os.makedirs("./system")
        with open("./system/blockMeshDict", "w") as f:
            # header
            f.writelines(
                [
                    "FoamFile\n",
                    "{\n",
                    "    version     2.0;\n",
                    "    format      ascii;\n",
                    "    class       dictionary;\n",
                    "    object      blockMeshDict;\n",
                    "}\n",
                    "\n",
                    "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n",
                    "\n",
                    "convertToMeters 1.0;\n",
                    "\n",
                ]
            )
            f.write("vertices\n(\n")
            for p in self.points:
                f.write(f"    ({p.x1} {p.x2} {p.x3})\n")
            f.write(");\n\n")
            f.write("edges\n(\n")
            for e in self.edges:
                if e.type == "line":
                    pass
                elif e.type == "arc":
                    f.write(f"    {e.type} {e.p0.id} {e.p1.id} (")
                    f.write(f"{e.points[0][0]} {e.points[0][1]} {e.points[0][2]}")
                    f.write(")\n")
                else:
                    f.write(f"    {e.type} {e.p0.id} {e.p1.id} (")
                    for p in e.points:
                        f.write(f" ({p[0]} {p[1]} {p[2]}) ")
                    f.write(")\n")
            f.write(");\n\n")
            f.write("blocks\n(\n")
            for b in self.blocks:
                f.write(
                    f"    hex ({b.p0.id} {b.p1.id} {b.p2.id} {b.p3.id} {b.p4.id} {b.p5.id} {b.p6.id} {b.p7.id})\n"
                )
                f.write(f"    ({b.cells_x1} {b.cells_x2} {b.cells_x3})\n")
                f.write(f"    {b.grading}\n")
            f.write(");\n\n")
            f.write("patches\n(\n")
            for p in self.patches:
                f.write(f"    {p.name}\n    (\n")
                for face in p.faces:
                    f.write(
                        f"    ({face[0].id} {face[1].id} {face[2].id} {face[3].id})\n"
                    )
                f.write("    )\n")
            f.write(");\n\n")
            f.write("mergePatchPairs\n(\n);\n\n")
            f.write(
                "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
            )


class Block:
    """Block of the mesh. Naming of points and edges following openFOAM standard."""

    def __init__(
        self,
        mesh,
        p0=[0, 0, 0],
        p1=[0, 0, 0],
        p2=[0, 0, 0],
        p3=[0, 0, 0],
        p4=[0, 0, 0],
        p5=[0, 0, 0],
        p6=[0, 0, 0],
        p7=[0, 0, 0],
    ) -> None:
        self.mesh = mesh
        self.id = -1

        self._p0_coords = p0
        self._p1_coords = p1
        self._p2_coords = p2
        self._p3_coords = p3
        self._p4_coords = p4
        self._p5_coords = p5
        self._p6_coords = p6
        self._p7_coords = p7

        self._cells_x1 = 0
        self._cells_x2 = 0
        self._cells_x3 = 0
        self.grading = "simpleGrading (1 1 1)"

        self._p0 = None
        self._p1 = None
        self._p2 = None
        self._p3 = None
        self._p4 = None
        self._p5 = None
        self._p6 = None
        self._p7 = None

        self.e0 = None
        self.e1 = None
        self.e2 = None
        self.e3 = None
        self.e4 = None
        self.e5 = None
        self.e6 = None
        self.e7 = None
        self.e8 = None
        self.e9 = None
        self.e10 = None
        self.e11 = None

        self._created = False

    def create(self):
        """Create points and edges"""
        if self._created:
            raise RuntimeError("This block was already crated.")
        if self._cells_x1 == 0 or self._cells_x2 == 0 or self._cells_x3 == 0:
            raise RuntimeError("Number of cells not defined.")

        if self._p0 is None:
            self._p0 = self.mesh._add_point(
                self._p0_coords[0],
                self._p0_coords[1],
                self._p0_coords[2],
            )
        if self._p1 is None:
            self._p1 = self.mesh._add_point(
                self._p1_coords[0],
                self._p1_coords[1],
                self._p1_coords[2],
            )
        if self._p2 is None:
            self._p2 = self.mesh._add_point(
                self._p2_coords[0],
                self._p2_coords[1],
                self._p2_coords[2],
            )
        if self._p3 is None:
            self._p3 = self.mesh._add_point(
                self._p3_coords[0],
                self._p3_coords[1],
                self._p3_coords[2],
            )
        if self._p4 is None:
            self._p4 = self.mesh._add_point(
                self._p4_coords[0], self._p4_coords[1], self._p4_coords[2]
            )
        if self._p5 is None:
            self._p5 = self.mesh._add_point(
                self._p5_coords[0],
                self._p5_coords[1],
                self._p5_coords[2],
            )
        if self._p6 is None:
            self._p6 = self.mesh._add_point(
                self._p6_coords[0], self._p6_coords[1], self._p6_coords[2]
            )
        if self._p7 is None:
            self._p7 = self.mesh._add_point(
                self._p7_coords[0], self._p7_coords[1], self._p7_coords[2]
            )

        if self.e0 is None:
            self.e0 = self.mesh._add_edge(self._p0, self._p1)
        if self.e1 is None:
            self.e1 = self.mesh._add_edge(self._p3, self._p2)
        if self.e2 is None:
            self.e2 = self.mesh._add_edge(self._p7, self._p6)
        if self.e3 is None:
            self.e3 = self.mesh._add_edge(self._p4, self._p5)
        if self.e4 is None:
            self.e4 = self.mesh._add_edge(self._p0, self._p3)
        if self.e5 is None:
            self.e5 = self.mesh._add_edge(self._p1, self._p2)
        if self.e6 is None:
            self.e6 = self.mesh._add_edge(self._p5, self._p6)
        if self.e7 is None:
            self.e7 = self.mesh._add_edge(self._p4, self._p7)
        if self.e8 is None:
            self.e8 = self.mesh._add_edge(self._p0, self._p4)
        if self.e9 is None:
            self.e9 = self.mesh._add_edge(self._p1, self._p5)
        if self.e10 is None:
            self.e10 = self.mesh._add_edge(self._p2, self._p6)
        if self.e11 is None:
            self.e11 = self.mesh._add_edge(self._p3, self._p7)
        self._created = True
        self.mesh._add_block(self)

    def set_connection(self, other, pos):
        """Set connection to other block. This works only if the
        coordinate systems have the same orientation!

        Args:
            other (Block): Block to set connection to.
            pos (str): Position of other block (respective to this one).
                       Options: top, bottom, left, right, front, back
        """
        if other._created == False:
            raise RuntimeError(
                "The other block was not created. Run other_block.create() first!"
            )
        if pos == "top":
            self._p4 = other._p0
            self._p5 = other._p1
            self._p6 = other._p2
            self._p7 = other._p3
            self.e3 = other.e0
            self.e2 = other.e1
            self.e6 = other.e5
            self.e7 = other.e4
            self._cells_x1 = other.cells_x1
            self._cells_x2 = other.cells_x2
        elif pos == "bottom":
            self._p0 = other._p4
            self._p1 = other._p5
            self._p2 = other._p6
            self._p3 = other._p7
            self.e0 = other.e3
            self.e1 = other.e2
            self.e5 = other.e6
            self.e4 = other.e7
            self._cells_x1 = other.cells_x1
            self._cells_x2 = other.cells_x2
        elif pos == "left":
            self._p0 = other._p1
            self._p3 = other._p2
            self._p4 = other._p5
            self._p7 = other._p6
            self.e4 = other.e5
            self.e7 = other.e6
            self.e8 = other.e9
            self.e11 = other.e10
            self._cells_x2 = other.cells_x2
            self._cells_x3 = other.cells_x3
        elif pos == "right":
            self._p1 = other._p0
            self._p2 = other._p3
            self._p5 = other._p4
            self._p6 = other._p7
            self.e5 = other.e4
            self.e6 = other.e7
            self.e9 = other.e8
            self.e10 = other.e11
            self._cells_x2 = other.cells_x2
            self._cells_x3 = other.cells_x3
        elif pos == "front":
            self._p0 = other._p3
            self._p1 = other._p2
            self._p4 = other._p7
            self._p5 = other._p6
            self.e0 = other.e1
            self.e3 = other.e2
            self.e8 = other.e11
            self.e9 = other.e10
            self._cells_x1 = other.cells_x1
            self._cells_x3 = other.cells_x3
        elif pos == "back":
            self._p3 = other._p0
            self._p2 = other._p1
            self._p7 = other._p4
            self._p6 = other._p5
            self.e1 = other.e0
            self.e2 = other.e3
            self.e11 = other.e8
            self.e10 = other.e9
            self._cells_x1 = other.cells_x1
            self._cells_x3 = other.cells_x3
        else:
            raise ValueError(
                "This position does not exist.\nThe following values are allowed for 'pos': top, bottom, left, right, front, back"
            )

    def set_number_of_cells(self, x1=10, x2=10, x3=10):
        self._cells_x1 = x1
        self._cells_x2 = x2
        self._cells_x3 = x3

    @property
    def cells_x1(self):
        return self._cells_x1

    @cells_x1.setter
    def cells_x1(self, val):
        if self._cells_x1 != 0:
            raise RuntimeError(
                "This value was already set or derived from a connected block."
            )
        self._cells_x1 = val

    @property
    def cells_x2(self):
        return self._cells_x2

    @cells_x2.setter
    def cells_x2(self, val):
        if self._cells_x2 != 0:
            raise RuntimeError(
                "This value was already set or derived from a connected block."
            )
        self._cells_x2 = val

    @property
    def cells_x3(self):
        return self._cells_x3

    @cells_x3.setter
    def cells_x3(self, val):
        if self._cells_x3 != 0:
            raise RuntimeError(
                "This value was already set or derived from a connected block."
            )
        self._cells_x3 = val

    @property
    def p0(self):
        return self._p0

    @p0.setter
    def p0(self, val):
        if self._p0 is not None:
            raise RuntimeError("This point exists already.")
        if type(val) == Point:
            self._p0 = val
        else:
            self._p0 = self.mesh._add_point(val[0], val[1], val[2])

    @property
    def p1(self):
        return self._p1

    @p1.setter
    def p1(self, val):
        if self._p1 is not None:
            raise RuntimeError("This point exists already.")
        if type(val) == Point:
            self._p1 = val
        else:
            self._p1 = self.mesh._add_point(val[0], val[1], val[2])

    @property
    def p2(self):
        return self._p2

    @p2.setter
    def p2(self, val):
        if self._p2 is not None:
            raise RuntimeError("This point exists already.")
        if type(val) == Point:
            self._p2 = val
        else:
            self._p2 = self.mesh._add_point(val[0], val[1], val[2])

    @property
    def p3(self):
        return self._p3

    @p3.setter
    def p3(self, val):
        if self._p3 is not None:
            raise RuntimeError("This point exists already.")
        if type(val) == Point:
            self._p3 = val
        else:
            self._p3 = self.mesh._add_point(val[0], val[1], val[2])

    @property
    def p4(self):
        return self._p4

    @p4.setter
    def p4(self, val):
        if self._p4 is not None:
            raise RuntimeError("This point exists already.")
        if type(val) == Point:
            self._p4 = val
        else:
            self._p4 = self.mesh._add_point(val[0], val[1], val[2])

    @property
    def p5(self):
        return self._p5

    @p5.setter
    def p5(self, val):
        if self._p5 is not None:
            raise RuntimeError("This point exists already.")
        if type(val) == Point:
            self._p5 = val
        else:
            self._p5 = self.mesh._add_point(val[0], val[1], val[2])

    @property
    def p6(self):
        return self._p6

    @p6.setter
    def p6(self, val):
        if self._p6 is not None:
            raise RuntimeError("This point exists already.")
        if type(val) == Point:
            self._p6 = val
        else:
            self._p6 = self.mesh._add_point(val[0], val[1], val[2])

    @property
    def p7(self):
        return self._p7

    @p7.setter
    def p7(self, val):
        if self._p7 is not None:
            raise RuntimeError("This point exists already.")
        if type(val) == Point:
            self._p7 = val
        else:
            self._p7 = self.mesh._add_point(val[0], val[1], val[2])

    @property
    def face_front(self):
        if not self._created:
            raise RuntimeError("This block was not created yet.")
        return [self._p0, self._p1, self._p5, self._p4]

    @face_front.setter
    def face_front(self, val):
        self._p1 = val[0]
        self._p0 = val[1]
        self._p4 = val[2]
        self._p5 = val[3]

    @property
    def face_back(self):
        if not self._created:
            raise RuntimeError("This block was not created yet.")
        return [self._p2, self._p3, self._p7, self._p6]

    @face_back.setter
    def face_back(self, val):
        self._p3 = val[0]
        self._p2 = val[1]
        self._p6 = val[2]
        self._p7 = val[3]

    @property
    def face_left(self):
        if not self._created:
            raise RuntimeError("This block was not created yet.")
        return [self._p3, self._p0, self._p4, self._p7]

    @face_left.setter
    def face_left(self, val):
        self._p0 = val[0]
        self._p3 = val[1]
        self._p7 = val[2]
        self._p4 = val[3]

    @property
    def face_right(self):
        if not self._created:
            raise RuntimeError("This block was not created yet.")
        return [self._p1, self._p2, self._p6, self._p5]

    @face_right.setter
    def face_right(self, val):
        self._p2 = val[0]
        self._p1 = val[1]
        self._p5 = val[2]
        self._p6 = val[3]

    @property
    def face_bottom(self):
        if not self._created:
            raise RuntimeError("This block was not created yet.")
        return [self._p2, self._p1, self._p0, self._p3]

    @face_bottom.setter
    def face_bottom(self, val):
        self._p3 = val[0]
        self._p0 = val[1]
        self._p1 = val[2]
        self._p2 = val[3]

    @property
    def face_top(self):
        if not self._created:
            raise RuntimeError("This block was not created yet.")
        return [self._p7, self._p4, self._p5, self._p6]

    @face_top.setter
    def face_top(self, val):
        self._p6 = val[0]
        self._p5 = val[1]
        self._p4 = val[2]
        self._p7 = val[3]


@dataclass
class Point:
    """Point in the mesh."""

    x1: float
    x2: float
    x3: float
    id: int = -1

    def update_coordinates(self, coords):
        self.x1 = coords[0]
        self.x2 = coords[1]
        self.x3 = coords[2]


class Edge:
    """Edge of a block."""

    def __init__(self, p0, p1) -> None:
        self.p0 = p0
        self.p1 = p1
        self.type = "line"
        self.points = []


class Patch:
    """Patch to define boundaries."""

    def __init__(self, mesh, name) -> None:
        self.name = name
        self.faces = []
        mesh.patches.append(self)

    def add_face(self, face):
        self.faces.append(face)


def boundary_layer(
    block_size,
    pos="xmin",
    smallest_element=0.0003,
    layer_thickness=0.007,
    growth_rate=1.2,
):
    """Compute boundary layer

    Args:
        block_size (float): length of block in direction of BL
        pos (str): Coordinate position for grading
        smallest_element (float): Minimum element size (preserved)
        layer_thickness (float): Thickness of BL (modified to match number of elements).
        growth_rate (float): Growth rate of element size in BL.

    Returns:
        number of elements, grading-string
    """
    if growth_rate < 1:
        growth_rate = 1 / growth_rate
    # compute number of elements in boundary layer
    if growth_rate == 1:
        n_el_bl = int(np.ceil(layer_thickness / smallest_element))
    else:
        n_el_bl = int(
            np.ceil(
                np.log(1 - layer_thickness * (1 - growth_rate) / smallest_element)
                / np.log(growth_rate)
            )
        )
    # update layer thickness to match rounded number of elements
    if growth_rate == 1:
        layer_thickness = smallest_element * n_el_bl
    else:
        layer_thickness = (
            smallest_element * (1 - growth_rate ** n_el_bl) / (1 - growth_rate)
        )
    # maximum element size
    if growth_rate == 1:
        largest_element = smallest_element
    else:
        largest_element = smallest_element * growth_rate ** (n_el_bl - 1)
    # number of elements outside of boundary layer
    n_el_out = int((block_size - layer_thickness) / largest_element)
    n_el = n_el_bl + n_el_out

    if pos == "xmin":
        grading = f"( ({layer_thickness / block_size} {n_el_bl} {growth_rate**(n_el_bl-1)}) ({(block_size-layer_thickness)/block_size} {n_el_out} 1) )"
    elif pos == "xmax":
        grading = f"( ({(block_size-layer_thickness)/block_size} {n_el_out} 1) ({layer_thickness/block_size} {n_el_bl} {1/growth_rate**(n_el_bl-1)}) )"
    else:
        raise ValueError(f"Position '{pos}'' not defined. Use either 'xmin' or 'xmax'")
    print("Elements in BL:", n_el_bl)
    print("Elements outside:", n_el_out)
    print("Layer thickness:", layer_thickness)
    print("Larges Element:", largest_element)
    return n_el, grading



"""Cylinders with an octagonal core. Does not provide the same functionality as cylinders.py."""
from dataclasses import dataclass
from .blocks import *


@dataclass
class Cylinder:
    inner_blocks: list
    outer_blocks: list
    surf_top: list
    surf_bt: list
    surf_rad: list


@dataclass
class Ring:
    blocks: list
    surf_top: list
    surf_bt: list
    surf_rad: list


def cartesian(r, phi, z, degree=True):
    if degree:
        phi = 2 * np.pi * phi / 360
    x = r * np.cos(phi)
    y = r * np.sin(phi)
    return [x, y, z]


def create_cylinder(
    mesh,
    r_top,
    r_bt,
    z_top,
    z_bt,
    z_top_mid,
    z_bt_mid,
    res_r,
    res_phi,
    res_z,
    s_in_top=[],
    s_in_bt=[],
    s_out_top=[],
    s_out_bt=[],
):
    # Central blocks
    radius_center_top = r_top / 2
    radius_center_bt = r_bt / 2

    res_center_r = int(res_phi / 8)
    b_0 = Block(
        mesh,
        cartesian(radius_center_bt, 0, z_bt_mid),
        cartesian(radius_center_bt, 45, z_bt_mid),
        cartesian(radius_center_bt, 90, z_bt_mid),
        cartesian(0, 0, z_bt_mid),
        cartesian(radius_center_top, 0, z_top_mid),
        cartesian(radius_center_top, 45, z_top_mid),
        cartesian(radius_center_top, 90, z_top_mid),
        cartesian(0, 0, z_top_mid),
    )
    b_0.set_number_of_cells(res_center_r, res_center_r, res_z)
    b_0.create()

    b_90 = Block(mesh)
    b_90.set_connection(b_0, "front")
    b_90.p2 = cartesian(radius_center_bt, 135, z_bt_mid)
    b_90.p3 = cartesian(radius_center_bt, 180, z_bt_mid)
    b_90.p6 = cartesian(radius_center_top, 135, z_top_mid)
    b_90.p7 = cartesian(radius_center_top, 180, z_top_mid)
    b_90.cells_x1 = res_center_r
    b_90.create()

    b_180 = Block(mesh)
    b_180.set_connection(b_90, "right")
    b_180.p0 = cartesian(radius_center_bt, 270, z_bt_mid)
    b_180.p3 = cartesian(radius_center_bt, 225, z_bt_mid)
    b_180.p4 = cartesian(radius_center_top, 270, z_top_mid)
    b_180.p7 = cartesian(radius_center_top, 225, z_top_mid)
    b_180.cells_x0 = res_center_r
    b_180.create()

    b_270 = Block(mesh)
    b_270.set_connection(b_180, "back")
    b_270.set_connection(b_0, "right")
    b_270.p0 = cartesian(radius_center_bt, 315, z_bt_mid)
    b_270.p4 = cartesian(radius_center_top, 315, z_top_mid)
    b_270.create()

    # if s_in_top != []:
    #     b_0.e7.type = "spline"
    #     b_90.e3.type = "spline"
    #     b_180.e6.type = "spline"
    #     b_270.e2.type = "spline"
    #     for p in s_in_top:
    #         b_0.e7.points.append(cartesian(p[0], 0, p[1]))
    #         b_90.e3.points.append(cartesian(p[0], 90, p[1]))
    #         b_180.e6.points.append(cartesian(p[0], 180, p[1]))
    #         b_270.e2.points.append(cartesian(p[0], 180, p[1]))

    # if s_in_bt != []:
    #     b_0.e4.type = "spline"
    #     b_90.e0.type = "spline"
    #     b_180.e5.type = "spline"
    #     b_270.e1.type = "spline"
    #     for i in range(len(s_in_bt)):
    #         coords = cartesian(s_in_bt[-i][0], 0, s_in_bt[-i][1])
    #         b_0.e4.points.append(cartesian(s_in_bt[-(i+1)][0], 0, s_in_bt[-(i+1)][1]))
    #         b_90.e0.points.append(cartesian(s_in_bt[i][0], 90, s_in_bt[i][1]))
    #         b_180.e5.points.append(cartesian(s_in_bt[i][0], 180, s_in_bt[i][1]))
    #         b_270.e1.points.append(cartesian(s_in_bt[-(i+1)][0], 270, s_in_bt[-(i+1)][1]))

    # #############
    # # TRY SPLINE INTERPOTALTION
    # x0 = np.array(cartesian(radius_center_bt, 0, z_bt_mid))
    # x1 = np.array(cartesian(radius_center_bt, 45, z_bt_mid))

    # radii = []
    # phis = []
    # res = 100
    # for dist in np.linspace(0, 1, res, endpoint=False):
    #     pos = x0 + dist*(x1-x0)
    #     radii.append((pos[0]**2 + pos[1]**2)**0.5)
    #     phis.append(np.arctan(pos[1]/pos[0])*360/(2*np.pi))
    # s_in_bt = [[0, 0]] + s_in_bt + [[radius_center_bt, z_bt_mid]]
    # s_in_bt = np.array(s_in_bt)
    # spline = interp1d(s_in_bt[:, 0], s_in_bt[:, 1], kind='cubic')
    # z_vals = spline(radii)

    # b_0.e0.type = "spline"
    # b_0.e5.type = "spline"
    # for i in range(res-1):
    #     b_0.e0.points.append(cartesian(radii[i+1], phis[i+1], z_vals[i+1]))
    #     b_0.e5.points.append(cartesian(radii[i+1], phis[i+1] + 45, z_vals[i+1]))

    # create ring around these blocks
    ring = create_ring(
        mesh,
        r_top,
        r_bt,
        z_top,
        z_bt,
        [
            b_0.face_front,
            b_0.face_right,
            b_90.face_right,
            b_90.face_back,
            b_180.face_back,
            b_180.face_left,
            b_270.face_left,
            b_270.face_front,
        ],
        res_r - res_center_r,
        res_phi,
        res_z,
    )

    surf_top = [
        b_0.face_top,
        b_90.face_top,
        b_180.face_top,
        b_270.face_top,
    ] + ring.surf_top
    surf_bt = [b_0.face_bottom, b_90.face_bottom, b_180.face_bottom, b_270.face_bottom]

    return Cylinder(
        [b_0, b_90, b_180, b_270], ring.blocks, surf_top, surf_bt, ring.surf_rad
    )


def create_ring(mesh, r_top, r_bt, z_top, z_bt, faces_inside, res_r, res_phi, res_z):

    blocks = []

    b = Block(mesh)
    b.face_left = faces_inside[0]
    b.p1 = cartesian(r_bt, 0, z_bt)
    b.p2 = cartesian(r_bt, 45, z_bt)
    b.p5 = cartesian(r_top, 0, z_top)
    b.p6 = cartesian(r_top, 45, z_top)
    b.set_number_of_cells(res_r, int(res_phi / 8), res_z)
    b.create()
    b.e5.type = "arc"
    b.e5.points = [cartesian(r_bt, 45 / 2, z_bt)]
    b.e6.type = "arc"
    b.e6.points = [cartesian(r_top, 45 / 2, z_top)]
    blocks.append(b)

    for i in range(6):
        i += 1
        b = Block(mesh)
        b.face_left = faces_inside[i]

        b.face_front = blocks[i - 1].face_back
        b.p2 = cartesian(r_bt, 45 * (i + 1), z_bt)
        b.p6 = cartesian(r_top, 45 * (i + 1), z_top)
        b.set_number_of_cells(res_r, int(res_phi / 8), res_z)
        b.create()
        b.e5.type = "arc"
        b.e5.points = [cartesian(r_bt, i * 45 + 45 / 2, z_bt)]
        b.e6.type = "arc"
        b.e6.points = [cartesian(r_top, i * 45 + 45 / 2, z_top)]
        blocks.append(b)

    b = Block(mesh)
    b.face_left = faces_inside[-1]
    b.face_front = blocks[-1].face_back
    b.face_back = blocks[0].face_front
    b.set_number_of_cells(res_r, int(res_phi / 8), res_z)
    b.create()
    b.e5.type = "arc"
    b.e5.points = [cartesian(r_bt, 360 - 45 / 2, z_bt)]
    b.e6.type = "arc"
    b.e6.points = [cartesian(r_top, 360 - 45 / 2, z_top)]

    surf_top = []
    surf_bt = []
    surf_rad = []
    for b in blocks:
        surf_top.append(b.face_top)
        surf_bt.append(b.face_bottom)
        surf_rad.append(b.face_right)

    return Ring(blocks, surf_top, surf_bt, surf_rad)



import setuptools
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nemoblock",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Arved Enders-Seidlitz",
    author_email="arved.enders-seidlitz@ikz-berlin.de",
    description="Utilities to generate blockMeshDicts for OpenFOAM.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nemocrys/nemoblock",
    packages=["nemoblock"],
    # include_package_data=True,
    # package_data={"": ["data/*.yml"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "matplotlib",
        "numpy",
        "scipy",
    ],
    python_requires=">=3.7",
)
